// -----------------------------------------------------------------------------
// this is a collection of helper functions for the ZZ -> 4 leptons analysis
// to: 
// - check if the decay is leptonic, 
// - check if the Zs are in the mass range we want, 
// - check if the events is fiducial, 
// - check if the leptons are form the same vertex, 
// - smear the x,y,z vertex of a track by the detector reolution
// ----------------------------------------------------------------------------

#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include "TRandom3.h"
#include "TLorentzVector.h"


using namespace std;

//------
// use enumartion: changing the channel order, the functions about the ZZ channel and name stay the same
enum ZZtype {e4=0, mu4, e2mu2};
//
int ZZchannel(vector<int> lepIDs) {
  int sumAbsIDs = 0;
  for (int i=0; i<4; i++) sumAbsIDs += fabs(lepIDs[i]);
  if (sumAbsIDs == 4*11) return ZZtype(e4);
  if (sumAbsIDs == 4*13) return ZZtype(mu4);
  if (sumAbsIDs == 2*11+2*13) return ZZtype(e2mu2);
}
//
string ZZname(int theType ) {
  string theName;
  if (theType == ZZtype(e4) ) theName="4e";
  if (theType == ZZtype(mu4) ) theName="4mu";
  if (theType == ZZtype(e2mu2) ) theName="2e2mu";
  return theName;
}
//------
 

//
bool isLeptonicZ (int id1,int id2){
  // check that #1 and #2 are particle + antiparticle, so consistent with Z as a parent
  // and also that they are either electrons or muons
  if ( id1+id2 == 0 && (abs(id1)==11 || abs(id1) == 13) ) return true;
  return false;
}

//
bool isLeptonicZZ(int id1, int id2, int id3, int id4){
  // check that #1 and #2 are from the same Z, as printout PYTHIA event printout suggests
  if ( isLeptonicZ(id1, id2) && isLeptonicZ(id3, id4) ) return true; 
  return false;
}

// overridden method (same name, different interface), taking the leptonIDs in a vector:
bool isLeptonicZZ(vector<int> vlepID){
  // check that #0 and #1 are from the same Z, and #2 and #3 from the other Z
  if ( isLeptonicZ(vlepID[0], vlepID[1]) && isLeptonicZ(vlepID[2], vlepID[3]) ) return true; 
  return false;
}

//
bool isOnShellZ(float mZ) {
  // mass cut to call it a Z
  if (	mZ > 66.0 &&  mZ < 116.0 ) return true;  
  return false;
}


//
bool isOnShellZZ(float mZ1, float mZ2) {
  // mass cuts to call it two Zs
  if (	isOnShellZ(mZ1) && isOnShellZ(mZ2) ) return true;  
  return false;
}

 

//
bool isFiducialZZ(vector<TLorentzVector> tleptons, vector<int> lepIDs) {
  // Check if the event is in the fiducial region , 
  // with no assumptions on the flavours, nor on the order of appearance of the leptons in the vectors
  int nPassEta=0;
  int nPassPT1=0;
  int nPassPT2=0;
  int nPassPT3=0;
  int nPassPT4=0;
  bool eta_OK = false;
  bool Pt_OK = false;
  bool DR_OK = false;
  bool mass_OK = false;

  float ETA_CUT = 2.7;
  float PT_CUT1 = 20.0;
  float PT_CUT2 = 15.0;
  float PT_CUT3 = 10.0;
  float PT_CUT4 =  5.0;

  for(int i=0;i<4;i++){
    if(fabs(tleptons[i].Eta()) < ETA_CUT) nPassEta++;
    if(tleptons[i].Pt() > PT_CUT1) nPassPT1++;
    if(tleptons[i].Pt() > PT_CUT2) nPassPT2++;
    if(tleptons[i].Pt() > PT_CUT3) nPassPT3++;
    if(tleptons[i].Pt() > PT_CUT4) nPassPT4++;
  }


  //check if there are at least 3 leptons in the event with Pt>10 , 2 with Pt>20, and 1 with Pt>20
  // and all above 7 GeV then check \DeltaR
  if(nPassEta==4) eta_OK = true;
  if(nPassPT1>=1 && nPassPT2>=2 && nPassPT3>=3 && nPassPT4==4) Pt_OK = true;

  // check the mass and DR
  // check all possiblities of pairing the leptons. If any option has both pairs with mass 66-116, we are OK with the masses

  vector<TLorentzVector> tZ1, tZ2; // a list of possible solutions for the dileptons forming the Z's
  int sumAbsIDs = 0;
  for (int i=0; i<4; i++) sumAbsIDs += fabs(lepIDs[i]);
  if (sumAbsIDs == 44 || sumAbsIDs == 52) { // 4e=4*11=44 or 4mu=4*13=52, but we pretend do not know the order they are placed in the vectors
    //
    // check DR>0.1 for ALL same flavour combinations, no matter what their chanrge is:
    DR_OK = true;
    for(int i=0; i<4; i++) { 
      for (int j=i+1; j<4; j++) { 
	if (tleptons[i].DeltaR(tleptons[j])  < 0.1) DR_OK = false; // All leptons are same Favlour: one failure is enough to destroy criterion
      }
    }
    //
    mass_OK = false; 
    int nPairings=0;
    if (lepIDs[0]*lepIDs[1] < 0) {nPairings++; tZ1.push_back(tleptons[0] + tleptons[1]);  tZ2.push_back(tleptons[2] + tleptons[3]); } // 01 23
    if (lepIDs[0]*lepIDs[2] < 0) {nPairings++; tZ1.push_back(tleptons[0] + tleptons[2]);  tZ2.push_back(tleptons[1] + tleptons[3]); } // 02 13
    if (lepIDs[0]*lepIDs[3] < 0) {nPairings++; tZ1.push_back(tleptons[0] + tleptons[3]);  tZ2.push_back(tleptons[1] + tleptons[2]); } // 03 12
    for (int i=0; i<nPairings; i++) {
      if ( isOnShellZZ(tZ1[i].M(), tZ2[i].M()) ) mass_OK = true;
    } // any pairing satisfying the criteria is fine.
  } 
  else {  // 2e2mu, but we pretend we do not know the order they are placed in the vectors
    DR_OK = false;
    mass_OK = false;

    if ( lepIDs[0]+lepIDs[1] ==0 ) { // 01 23
      if (tleptons[0].DeltaR(tleptons[1]) >=0.1     && tleptons[2].DeltaR(tleptons[3]) >=0.1
	  && tleptons[0].DeltaR(tleptons[2]) >=0.2 && tleptons[0].DeltaR(tleptons[3]) >=0.2
	  && tleptons[1].DeltaR(tleptons[2]) >=0.2  && tleptons[1].DeltaR(tleptons[3]) >=0.2 ) DR_OK = true; 
      if ( isOnShellZZ((tleptons[0] + tleptons[1]).M() , (tleptons[2] + tleptons[3]).M() ) ) mass_OK = true;
    } // end 01, 23

    if  (lepIDs[0]+lepIDs[2] == 0) { // 02 13
      if (tleptons[0].DeltaR(tleptons[2]) >=0.1     && tleptons[1].DeltaR(tleptons[3]) >=0.1
	  && tleptons[0].DeltaR(tleptons[1]) >=0.2 && tleptons[0].DeltaR(tleptons[3]) >=0.2
	  && tleptons[1].DeltaR(tleptons[2]) >=0.2  && tleptons[2].DeltaR(tleptons[3]) >=0.2 ) DR_OK = true; 
      if ( isOnShellZZ((tleptons[0] + tleptons[2]).M() , (tleptons[1] + tleptons[3]).M() ) ) mass_OK = true; 
    }

    if  (lepIDs[0]+lepIDs[3] == 0) { // 03 12
      if (tleptons[0].DeltaR(tleptons[3]) >=0.1     && tleptons[1].DeltaR(tleptons[2]) >=0.1
	  && tleptons[0].DeltaR(tleptons[1]) >=0.2 && tleptons[0].DeltaR(tleptons[2]) >=0.2
	  && tleptons[1].DeltaR(tleptons[3]) >=0.2  && tleptons[2].DeltaR(tleptons[3]) >=0.2 ) DR_OK = true; 
      if( isOnShellZZ((tleptons[0] + tleptons[3]).M() , (tleptons[1] + tleptons[2]).M() ) ) mass_OK = true; 
    } // end 03 12

  } // check masses and DRs for all pairings. Any combination passing the mass criteria is fine

  // if (mass_OK) {
  if (eta_OK && Pt_OK && DR_OK && mass_OK) {
    return true;
  } else {
    return false;
  }
}



//
bool isConsistentWithSameVertex(double zpp_A, double zpp_B, vector<double> lep_vz, vector<double> lep_sintheta, vector<double> lep_pt, bool printout=false) {
  // Check if all 4 leptons are consistent with comming from the same point on the z-axis
  //
  float z0SinTheta_CUT = 0.5; // cut value used in the analysis

  int nOK_A=0;  // number of leptons consistent with the z-vertex A
  int nOK_B=0;  // number of leptons consistent with the z-vertex B
  int nOK_ave4l=0;  // number of leptons consistent with their weighted average z0 positions
  int nLeptons = lep_vz.size();
  
  double zpp_ave4l = 0; // reco-like primary Z calulation = average z position, weighted by the Pt^2 
  double sumPt2 = 0;
  for (int i=0; i<nLeptons; i++) {
    sumPt2 += lep_pt[i] * lep_pt[i];
    zpp_ave4l += lep_vz[i]*(lep_pt[i] * lep_pt[i]);
  }
  zpp_ave4l /= sumPt2; // reco-like primary Z calulation = average z position, weighted by the Pt^2
  
  for (int i=0; i<nLeptons; i++) {
    if ( abs( (lep_vz[i] - zpp_A)*lep_sintheta[i] ) < z0SinTheta_CUT ) nOK_A++;
    if ( abs( (lep_vz[i] - zpp_B)*lep_sintheta[i] ) < z0SinTheta_CUT ) nOK_B++;
    if ( abs( (lep_vz[i] - zpp_ave4l)*lep_sintheta[i] ) < z0SinTheta_CUT ) nOK_ave4l++;
  }

  if (nOK_A == nLeptons || nOK_B == nLeptons || nOK_ave4l == nLeptons) {
    if (printout) {
      cout << nLeptons << " leptons found: " 
	   << nOK_A << " leptons consistent with the first vertex at " << zpp_A << " and " 
	   << nOK_B << " consistent with the second vertex at " << zpp_B << " and " 
	   << nOK_ave4l<< " consistent with their weighted average z0 " << zpp_ave4l
	   << endl;
    }
    return true;
  }

  return false;
}


//
void recoXYZ(unsigned long mySeed, double v0XYZ[3], double theta0, double phi0, vector<double> &vXYZ) {
  // Given a randomNumberSeed and the generated {x0,y0,z0,theta0,phi0} of a track, return the {X,Y,Z} moved randomly by resolution
  //
  // https://twiki.cern.ch/twiki/bin/view/AtlasPublic/InDetTrackingPerformanceApprovedPlots#Run_2    <-- see resolutiosn etc there
  // assume some resolution: https://atlas.web.cern.ch/Atlas/GROUPS/PHYSICS/PAPERS/STDM-2015-02/
  //
  float d0_sigma         = 0.2; // cut value used in the analysis
  float z0SinTheta_sigma = 0.2; // cut value used in the analysis

  TRandom3 rnd;
  rnd.SetSeed(mySeed); // new seed for each particle we treat
  vXYZ.clear();
  vXYZ.push_back( rnd.Gaus( v0XYZ[0], d0_sigma*cos(phi0) ) ); // assume sigma_d0 ~ 0.2 mm , x0 = d0*cos(phi)
  vXYZ.push_back( rnd.Gaus( v0XYZ[1], d0_sigma*sin(phi0) ) ); // assume sigma_d0 ~ 0.2 mm , y0= d0*sin(phi)
  double vzsintheta = rnd.Gaus( v0XYZ[2]*sin(theta0), z0SinTheta_sigma); // sigma_z0 * sin(theta) ~ 0.2mm
  vXYZ.push_back( vzsintheta/sin(theta0) );
  return;
}

