// Header file to access Pythia 8 program elements.
#include "Pythia8/Pythia.h"
#include <cmath>
#include <iostream>
#include <stdexcept>
#include <string>
#include "TTree.h"
#include "TFile.h"
#include "TH1.h"
#include "TCanvas.h"
#include "TLorentzVector.h"

using namespace Pythia8;

//
bool isLeptonicZ (int id1,int id2){
  // check that #1 and #2 are particle + antiparticle, so consistent with Z as a parent
  // and also that they are either electrons or muons
  if ( id1+id2 == 0 && (abs(id1)==11 || abs(id1) == 13) ) return true;
  return false;
}

//
bool isOnShellZ(float mZ) {
  // mass cut to call it a Z
  if (  mZ > 66.0 &&  mZ < 116.0 ) return true;
  return false;
}



int main(int argc, char * argv[]){
  
  // Create Output file and TTrees to be stored in there 
  TFile *outputfile = TFile::Open("zpythia.root","recreate");

  // Structure with individual values:
  struct resultsStruct{
    Int_t pidI1;
    Int_t pidI2;
    Int_t pidL1;
    Int_t pidL2;
    Float_t prodX;
    Float_t prodY;
    Float_t prodZ;
    Float_t mz;
    Float_t m12; // the correct lepton combination from Z1  (PYTHIA outputs the leptons in this order: 1,2,3,4 = -,+,-,+)
    Float_t pt12;
    Float_t dphi12; 
  };
  resultsStruct myStruct;

  TLorentzVector tlep1, tlep2, t12; // will be needed later: to be over-written in each event

  // Vectors with particle info: pythia_statusCode, pdgID and complete 4-vector allows any analysis to be done afterwards:
  std::vector<Int_t>     mc_pid; // PDG id for MonteCarlo particles
  std::vector<Int_t>     mc_status; // PYTHIA status code
  std::vector<Double_t>   mc_E; // self explanatory
  std::vector<Double_t>   mc_px;
  std::vector<Double_t>   mc_py;
  std::vector<Double_t>   mc_pz;
  std::vector<Double_t>   mc_sintheta;
  std::vector<Double_t>   mc_vx;
  std::vector<Double_t>   mc_vy;
  std::vector<Double_t>   mc_vz;


  // Create the  TTrees to store in the file:
  TTree *mcInfoTree = new TTree("mcInfo","MC inf Tree"); 
  mcInfoTree->Branch("mcValues",&myStruct,
		     "pidI1/I:pidI2/I:pidL1/I:pidL2/I:prodX/F:prodY/F:prodZ/F:mz/F:m12/F:pt12/F:dphi12/F");
  

  TTree *leptonsTree = new TTree("leptons","leptons");
  leptonsTree->Branch("mc_pid",    &mc_pid );
  leptonsTree->Branch("mc_status", &mc_status);
  leptonsTree->Branch("mc_E",      &mc_E);
  leptonsTree->Branch("mc_px",     &mc_px);
  leptonsTree->Branch("mc_py",     &mc_py);
  leptonsTree->Branch("mc_pz",     &mc_pz);
  leptonsTree->Branch("mc_sintheta",  &mc_sintheta);
  leptonsTree->Branch("mc_vx",     &mc_vx);
  leptonsTree->Branch("mc_vy",     &mc_vy);
  leptonsTree->Branch("mc_vz",     &mc_vz);


  // Generator and shorthand for the event	
  Pythia pythia;

  // Read in commands from external file
  pythia.readFile(argv[1]);

  // Initialize
  pythia.init();
  Event & event = pythia.process;

  // Extract settings to be used in the main program.
  int nEvent  = pythia.mode("Main:numberOfEvents");
       
  // int nAbort  = pythia.mode("Main:timesAllowErrors");

 
  // Begin event loop. Generate event; skip if generation aborted, or event not interesting:
  //
  unsigned long nTimesInRandom =0; // count how many times we call the TRandom3 and use this as a seed
  Int_t nLeptonic =0; // number of events where both Z's decay to electrons or muons
  float nPrints = 10.;
  int nPrintsLeft = nPrints;
  int everyNthEvent = int( (1.0*nEvent)/nPrints ); // split the sample so that you print 10 times
  cout << " Print the event record every "  << everyNthEvent << " event" << endl;

  for (int iEvent = 0; iEvent < nEvent ; ++iEvent) {
    if (!pythia.next()) continue;

    //clear vectors
    mc_pid.clear();
    mc_status.clear();
    mc_E.clear();
    mc_px.clear();
    mc_py.clear();
    mc_pz.clear();
    mc_sintheta.clear();
    mc_vx.clear();
    mc_vy.clear();
    mc_vz.clear();
		
    // The pythia stats() printout gives us the particles in the hard process: 
    // 3 + 4 -> 5  -> 6 + 7 ( parton + parton -> Z -> lepton1 + lepton2 ) 
    Particle const & init1 = event[3];  // each event[i] is a pythia Particle object
    Particle const & init2 = event[4];
    Particle const & Z1 = event[5];
    Particle const & lep1 = event[6];
    Particle const & lep2 = event[7];

    // check if event is interesting: leptons are particle + antiparticle, and only electrons or muons
    if ( !( lep1.id()+lep2.id() == 0 && (abs(lep1.id())==11 || abs(lep1.id()) == 13) ) ) continue; // try next event
    //if  ( !isLeptonicZ(lep1.id(), lep2.id() ) ) continue; // try next event using the check above implemented in a bool method
    if ( !isOnShellZ( Z1.m() ) ) continue; // it's not an on-shell Z: try next event (check is implemented in a bool method defined just after the includes))
    nLeptonic++;

    // First fill the mcInfo TTree:
    // ----------------------------
    myStruct.pidI1 = init1.id();
    myStruct.pidI2 = init2.id();
    myStruct.pidL1 = lep1.id();
    myStruct.pidL2 = lep2.id();
    myStruct.prodX = Z1.xProd();
    myStruct.prodY = Z1.yProd();
    myStruct.prodZ = Z1.zProd();
    myStruct.mz    = Z1.m();  // m() is the method of PYTHIA for the mass of a PYTHI Particle
    tlep1.SetPxPyPzE( lep1.px(), lep1.py(), lep1.pz(), lep1.e() );  // the classic constructor of TLorentzVector
    tlep2.SetPxPyPzE( lep2.px(), lep2.py(), lep2.pz(), lep2.e() );
    t12=tlep1+tlep2;
    myStruct.m12 = t12.M();  // the method of TLorentzVector for the mass is M()
    myStruct.pt12 = t12.Pt(); 
    myStruct.dphi12 = tlep1.DeltaPhi(tlep2); // Delta i between the lepton 1 and 2
    
    mcInfoTree->Fill();



    // Now fill the leptons TTree:
    // ---------------------------

    // 1st initial parton: #0
    mc_pid.push_back( init1.id() );
    mc_status.push_back( init1.status() );
    mc_E.push_back( init1.e() );
    mc_px.push_back( init1.px() );
    mc_py.push_back( init1.py() );
    mc_pz.push_back( init1.pz() );
    mc_sintheta.push_back( sin(init1.theta()) );
    mc_vx.push_back( init1.xProd() );
    mc_vy.push_back( init1.yProd() );
    mc_vz.push_back( init1.zProd() );
	  
    // 2nd initial parton: #1
    mc_pid.push_back( init2.id() );
    mc_status.push_back( init2.status() );
    mc_E.push_back( init2.e() );
    mc_px.push_back( init2.px() );
    mc_py.push_back( init2.py() );
    mc_pz.push_back( init2.pz() );
    mc_sintheta.push_back( sin(init2.theta()) );
    mc_vx.push_back( init2.xProd() );
    mc_vy.push_back( init2.yProd() );
    mc_vz.push_back( init2.zProd() );

    // 1st lepton: #2
    mc_pid.push_back( lep1.id() );
    mc_status.push_back( lep1.status() );
    mc_E.push_back( lep1.e() );
    mc_px.push_back( lep1.px() );
    mc_py.push_back( lep1.py() );
    mc_pz.push_back( lep1.pz() );
    mc_sintheta.push_back( sin(lep1.theta()) );
    mc_vx.push_back(lep1.xProd() );
    mc_vy.push_back(lep1.yProd() );
    mc_vz.push_back(lep1.zProd() );

    // 2nd lepton: #3
    mc_pid.push_back( lep2.id() );
    mc_status.push_back( lep2.status() );
    mc_E.push_back( lep2.e() );
    mc_px.push_back( lep2.px() );
    mc_py.push_back( lep2.py() );
    mc_pz.push_back( lep2.pz() );
    mc_sintheta.push_back( sin(lep2.theta()) );
    mc_vx.push_back(lep2.xProd() );
    mc_vy.push_back(lep2.yProd() );
    mc_vz.push_back(lep2.zProd() );


    //Sumpx py pz vectors: one for initial state, and one for final state
    // I use pid =0 (non physicsal value) for the the Sum of the 4-vectos: before and after

    // Sum of initial state: #4
    mc_pid.push_back(0);  
    mc_status.push_back(0);
    mc_E.push_back( init1.e() + init2.e() );
    mc_px.push_back( init1.px() + init2.px()  );
    mc_py.push_back( init1.py() + init2.py() );
    mc_pz.push_back( init1.pz() + init2.pz()  );
    mc_sintheta.push_back( sin(init1.theta()) );
    mc_vx.push_back( init1.xProd() );
    mc_vy.push_back( init1.yProd() );
    mc_vz.push_back( init1.zProd() );


    // Sum of final state: #5
    mc_pid.push_back(0);  // pid =0 means for me that I look at the Sum of the 4-vectos: before and after
    mc_status.push_back(0);
    mc_E.push_back( lep1.e() + lep2.e() );
    mc_px.push_back( lep1.px() + lep2.px() );
    mc_py.push_back( lep1.py() + lep2.py() );
    mc_pz.push_back( lep1.pz() + lep2.pz() );
    mc_sintheta.push_back( sin(Z1.theta()) );
    mc_vx.push_back( Z1.xProd() );
    mc_vy.push_back( Z1.yProd() );
    mc_vz.push_back( Z1.zProd() );
				
    leptonsTree->Fill();			
		
    // Print 1st and every Nth leptonic event, to see that I got them right:
    if ( nLeptonic==1 || nLeptonic%everyNthEvent== 0 ) {  // print 1st and every Nth leptonic event, to see that I got them right:
      cout << " *****" << endl;
      if (nLeptonic != 1) nPrintsLeft--; // do not count the fist time. But the rest yes. 
      cout << " ***** Leptonic event number: " << nLeptonic << " : " << 100.*nPrintsLeft/nPrints << " \% still left to do " << endl;
      cout << " *****" << endl;
      pythia.process.list();

      cout << "------- PRODUCTION point for each particle (x,y,z) mm: " << endl;
      cout << "z1: "   << mc_vx[0] << " , " << mc_vy[0] << " , " << mc_vz[0] << endl;
      cout << "lep1: "   << mc_vx[2] << " , " << mc_vy[2]  << " , " << mc_vz[2] <<  " ;  DetlaZ*sinTheta :" << (mc_vz[2] - mc_vz[0])*mc_sintheta[2] << endl;
      cout << "lep2: "   << mc_vx[3] << " , " << mc_vy[3]  << " , " << mc_vz[3] <<  " ;  DetlaZ*sinTheta :" << (mc_vz[3] - mc_vz[0])*mc_sintheta[3] << endl;
      cout << "------- " << endl;
    }

  } //end for loop

  // Sample statistics: 
  cout << "***** nLeptonic / nEvents = " << nLeptonic << " / " << nEvent << " = (" << 100.0*nLeptonic/nEvent 
       << " +- " << 100.0* sqrt(nLeptonic)/nEvent << ") \%"  << endl;
  cout << "***** expected from the 3.36\% BR to just Z -> ee or just Z -> mumu = 2* 3.36\% = " 
       << 100 * 2*(3.36/100) << " \%" << endl;  

  outputfile->Write();
  outputfile->Close();
	
  // delete whatever I had created with "new" to make sure the memory there gets free:
  // delete mcInfoTree;
  // delete leptonsTree;

  // print cross sections etc:
  pythia.statistics();
	
  return 0;  // return an integer, as the declation of the main() wants
}
